############################################################
#                                                          #
#              Information                                 #
#                                                          #
############################################################

# Dates: 1/01/2021 till 30/12/2021 We dont have data for 31 Dec. and the code don't resist it. We need more robustness   
# Time zone and hour format: Hour of the day is shown in UTC. Value at Hour 1 correspond to the hour starting at 1 and ending at 2. As April falls in summer time (UTC + 2 hours), to convert from local time to UTC we move two hours backwards. So, 7 am local time is 5 am UTC.   
# Pollutant: NO2   
# Catalunya    
# Sites: 72 monitoring stations
# Observations source: 
# Models intercompared:  
# Raw model from Caliope
# LOOCV applied to Caliope
#-------------------------------------------------------------------------------
# Measure the time it takes to run the analysis
start_time <- Sys.time()
#-------------------------------------------------------------------------------
#              LOOCV TO CALIOPE DATA WITH CATALONIA OBSERVATIONS
#-------------------------------------------------------------------------------

############################################################
#                                                          #
#              Set libraries                               #
#                                                          #
############################################################
# ! check the version of the package or R if any library or function does not work correctly

library(urbanR)
library(RNetCDF)
library(sp)
library(sf)
#library(mapview)
library(ncdf4)
library(rgeos) 
library(ggplot2)
library(plyr) #OJO! la llibreria plyr ha de ser cridada abans que dplyr 
library(dplyr)
library(broom)
library(magrittr)
library(ggpubr)
library(raster)
library(rgdal)
#library(tmap)
library(gstat)
library(sp)
library(abind)
library(openair)
#library(Metrics)

############################################################
#                                                          #
#              LOAD config_file                            #
#                                                          #
############################################################
print("starting")

args = commandArgs(TRUE)
if (length(args) > 0) {
  configFile_path <- args[1]
} else {
  cat("\n***WARNING: no ConfigFile was provided. Using default ConfigFile.\n")
}

#config_file it's a file where we change the data, where we want to save the results, etc. so now we extract these data from it:
#configFile_path <- "/esarchive/scratch/mhajji/Meriem/RUN_UK_regional/LOOCV/config_file_2021.yml"
#configFile_path <- "/esarchive/scratch/jmateu/Codes/run_geostats/config_file_2.yml"
config          <- yaml::yaml.load_file(configFile_path)
sim_folder_path <- config$sim_folder_path
obs_path        <- config$obs_path
model_path      <- config$model_path
save_result     <- config$save_result

date_time_end   <- config$date_time_end
date_time_begin <- config$date_time_begin

name            <- config$doc_name
variogram_path  <- config$variogram_path
path_code       <- config$path_code

set_stations       <- config$set_stations
set_observations   <- config$set_observations 
set_model          <- config$set_model
execute_loocv      <- config$execute_loocv
execute_statistics <- config$execute_statistics

save_general_statistics    <- config$save_general_statistics #Images
save_time_series           <- config$save_time_series #Images 1
save_temporal_variation    <- config$save_temporal_variation #Imatges 2


# As we had problems with Ordinary and Universal Kriging variograms, we saved one of each to read them in case we have problems with the one executed for a specific date or hour
variogram_ok <- paste0(variogram_path, "Variogram_OK.rds")
variogram_uk <- paste0(variogram_path, "Variogram_UK.rds")

compute_in_UTM        <- TRUE #We want to transform the coordinates from CSR to UTM, it is faster in Universal Kriging
crs_model_utm         <- "+proj=utm +zone=31 +ellps=intl +units=m +no_defs"
crs_model             <- "+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0" #coordinates sistem
############################################################
#                                                          #
#              Set temporal context                        #
#                                                          #
############################################################
print("setting dates")

#date_times et fa un super vector d'una  columna, dia per dia, hora per hora des de l'inici fins a la data final p.ex: 2021-01-01 00:00:00
date_times <- data_frame_dates(date_time_begin = date_time_begin,
                               date_time_end = date_time_end)

#dates: Llista de caracters anymesdia p.ex: 20210101
dates <- unique(paste0(as.character(strftime(date_times[, 1], format="%Y", tz = "UTC")),
                       as.character(strftime(date_times[, 1], format="%m", tz = "UTC")),
                       as.character(strftime(date_times[, 1], format="%d", tz = "UTC"))))
#year_month: igual que date pero anymes p.ex: 202101
year_month <- unique(paste0(as.character(strftime(date_times[, 1], format="%Y", tz = "UTC")),
                            as.character(strftime(date_times[, 1], format="%m", tz = "UTC"))))

############################################################
#                                                          #
#                 Load Local Stations                      #
#                                                          #
############################################################
print("Load Local Stations")

# local_stations: special points data frame with two cols [NO2 and ref_obs-reference of station-], at the moment we have 72 stations in Cat

if (set_stations){
  source(paste0(path_code,"set_stations.R"))
  print("Set stations load: done")
} else {
  local_stations       <- readRDS(file = paste0(sim_folder_path,"local_stations.rds"))
  print("Read local stations: done")
}

############################################################
#                                                          #
#                 Load Observations                        #
#                                                          #
############################################################
print("Load observations")
#we have a doc where we have the reference of the stations that are in catalonia, if we want to change or amplify them we can change the coordinates of the box in set_stations script

pollutant      <-"no2"

# we prepare a matrix where we will save the observations:
obs                  <- matrix(NA,ncol=(nrow(local_stations@data)),nrow=0)
df_obs_all           <- data.frame(obs)
dates_all            <- data.frame(matrix(NA,ncol= 1,nrow=0))
#we transform from a matrix to a data frame, with the reference of the station as a name of the col
colnames(df_obs_all) <- c(local_stations@data$ref_obs)
colnames(dates_all)  <- c("date")

# finally we extract the observations of certain year_month
for (y_month in year_month) {
  #obs_path<-"/esarchive/scratch/mhajji/Meriem/RUN_UK_regional/LOOCV/observacions/"
  netcdf_obs    <- RNetCDF::open.nc(paste0(obs_path, "sconc", pollutant, "_", y_month, ".nc"))

  code_obs      <- RNetCDF::var.get.nc(netcdf_obs, 'station_reference')
  #station_name  <- RNetCDF::var.get.nc(netcdf_obs, "station_name")
  obs_all       <- RNetCDF::var.get.nc(netcdf_obs, "sconcno2")
  
  year            <- as.numeric(substr(y_month, 1, 4))
  month           <- as.numeric(substr(y_month, 5, 6))
  num_days        <- length(RNetCDF::var.get.nc(netcdf_obs, "time"))/24
  date_times_loc  <- data_frame_dates( c(year,month,01,00) , c(year,month,num_days,23) )
	dates_all       <- rbind(dates_all,date_times_loc)

  pos_local_stations  <- which(code_obs%in%local_stations@data$ref_obs)
  cat_obs             <- obs_all[,pos_local_stations]
  cat_ref             <- code_obs[pos_local_stations]
  #cat_names           <- station_name[pos_local_stations]
  
  obs_month               <- matrix(NA,ncol=length(cat_obs[1,]),nrow=length(cat_obs[,1])) 
  #cat_obs matriu de 72 col (estacions) i en aquest cas p.ex: 744 (31 dies de generx24h) i fem una obs month buida
  df_obs_month            <- data.frame(obs_month)
  colnames(df_obs_month)  <- paste0(cat_ref)
  df_obs_month[,]         <- cat_obs*1.88
  #df_obs_month            <- cbind(date = date_times_loc$date, df_obs_month)
  
 
    df_obs_all              <- rbind.fill(df_obs_all, df_obs_month) #in case there is no data on a certain date, it fills it with NA (rbind.fill)
    if ( ncol(df_obs_all)!=ncol(df_obs_month) ) {
      print(paste0("ERROR: number of monitoring stations is not constant when loading observations. ","Filled with NA in: ",y_month))
    }
}

df_obs_all <- cbind(dates_all,df_obs_all)

saveRDS(object=df_obs_all, file=paste0(save_result,"/df_obs.rds"))

#time_subset permits us to select the observations at the same dates as the model (remember: at first we extract observations monthly)
df_obs_sub <- time_subset(data = df_obs_all, time_var_name = "date",
                          date_time_begin = date_time_begin,
                          date_time_end = date_time_end)

############################################################
#                                                          #
#                 Load Model                               #
#                                                          #
############################################################
print("Load Model")
var_name_reduc       <- "sconcno2" #pollutant
n_lat                <- 278        #latitude
n_lon                <- 298        #longitude
model                <- array(dim=c(lat=n_lat,lon=n_lon,hour=0)) #we prepare an array to put the date in

# We download the model by days
for (date in dates) {
  
  #date<-"20210101"
  conc_path_day  <- paste0(date, "00")
  file           <- paste(var_name_reduc, "_", conc_path_day, ".nc", sep="")
  file_path      <- paste0(model_path, file)
  
  if (file.exists(file_path)) {
    netcdf_model      <- RNetCDF::open.nc(file_path) #Dintre de la llibreria executem la funció open i aplica-ho al fitxer obs
    
    #Extraiem les dades del model
    model_day      <- RNetCDF::var.get.nc(netcdf_model, "sconcno2") #Concentracions del NO2
    model_day      <- model_day[,,1:24]
    model_day      <- model_day*1880 #les dades del model les tens en ppm i nosaltres ho volem en microg/m3
  }else{
    model_day       <- array(NaN, dim=c(n_lat,n_lon,24)) #if we dont have data for a certain day its filled by NaN
  }
  model            <- abind(model,model_day,along=3)
}
# aqui
lat_model      <- RNetCDF::var.get.nc(netcdf_model, 'lat') #Latitut
lat_model      <- sapply(lat_model, as.numeric)            #era una large matrix, ara es una large numeric
lon_model      <- RNetCDF::var.get.nc(netcdf_model, 'lon') #Longitud
lon_model      <- sapply(lon_model, as.numeric)

saveRDS(object=model, file=paste0(save_result,"/model.rds"))
        

############################################################
#                                                          #
#           Interpolate Model to stations' locations       #
#                                                          #
############################################################
print("OK to stations' locations")


#LOOCV_total and llista_uk_total is created empty to fill it later with LOOCV and raw model data respectivelly
ok_total           <- matrix(NA,ncol=nrow(local_stations@data),nrow=0) #list(length=nrow(date_times)) #we will have a large list with a data.frame for each hour
colnames(ok_total) <- local_stations$ref_obs #we want a matrix with empty data and the name of stations

#starting OK and LOOCV:
for (hour in 1:nrow(date_times)) {
  print(paste0("loop OK number:",hour))
  
  df_model <- model[,,hour]
  
  #if all the model for a certain hour is nan we cannot apply UK
  if ( sum(is.nan(df_model))!=length(df_model)  ){
    
    #observations:
    df_obs   <- df_obs_sub[hour,-1]
    df_obs                     <- data.frame(lat=local_stations@coords[,2],lon=local_stations@coords[,1],NO2=as.numeric(df_obs[1,]),code=colnames(df_obs))
    coordinates(df_obs)        <- ~lon+lat
    sp::proj4string(df_obs)    <- crs_model_utm    
    # 1. Interpolate model values at obs locations
    #    1.1. Set spatialpointsdataframe
    #raw model:
    df_model                  <- data.frame(lat=lat_model,lon=lon_model,NO2=sapply(df_model,as.matrix))
    coordinates(df_model)     <- ~lon+lat
    sp::proj4string(df_model) <- crs_model
    
    #config the coordinates in UTM
    df_model <- spTransform(df_model, sp::CRS(crs_model_utm))
    
    
    
    
    # 1.2 Compute Variogram
    set.seed(1) #we are picking random points in the next line, set.seed permits us to have same points in each loop
    pick_random_points  <- sample(1:nrow(df_model),2000)
    NO2.fit_ok          <- automap::autofitVariogram(formula=NO2~1, input_data=df_model[pick_random_points,],model="Ste")
    # Hem posat model='Ste' pk la segona hora del 01/06/2021 peta el variograma, mirar bé perque no s'ajusa
    
    #    1.3 Compute ordinary Kriging
    
    #Del model interpolo cap a les coordenades de les obs 'OK'
    ok_gstat <- gstat(formula=NO2~1, data = df_model, model = NO2.fit_ok$var_model, nmax=100)
    
    #Interpolacio del model a les coordenades del conjunt
    OK       <- predict(object=ok_gstat, newdata=df_obs)
    
    #If the variogram has singular matrix then it produces NAN's, so in that case we take a good Variogram instead:
    if(any(is.na(OK@data$var1.pred))){
      print("WARNING: VARIOGRAM DIVERGED! Using a precalculated VARIOGRAM")
      NO2.fit_ok=readRDS(variogram_ok)
      ok_gstat<- gstat(formula=NO2~1, data = df_model, model = NO2.fit_ok$var_model, nmax=100)
      
      #Interpolacio del model a les coordenades del conjunt
      OK <-predict(object=ok_gstat, newdata=df_obs)
    }
    
    
    # Save LOOCV and raw_model
    ok_h           <- matrix(data=OK@data$var1.pred, nrow=1)
    colnames(ok_h) <- df_obs$code
    
    
    ok_total <-  plyr::rbind.fill(as.data.frame(ok_total), as.data.frame(ok_h))
    
    
    
  }else{
    ok_h           <- matrix(NA,ncol=nrow(local_stations@data),nrow=1) 
    colnames(ok_h) <- local_stations$ref_obs
    
    ok_total           <- plyr::rbind.fill(as.data.frame(ok_total),as.data.frame(ok_h))
  }
  
}


############################################################
#                                                          #
#                 Loop Universal Kriging                   #
#                                                          #
############################################################
print("LOOCV")
#LOOCV_total and llista_uk_total is created empty to fill it later with LOOCV and raw model data respectively
LOOCV_total_1row           <- matrix(NA,ncol=nrow(local_stations@data),nrow=0) #list(length=nrow(date_times)) #we will have a large list with a data.frame for each hour
colnames(LOOCV_total_1row) <- local_stations$ref_obs #we want a matrix with empty data and the name of stations

#
hour_vector <- 1:nrow(date_times)
dim(hour_vector) <- c(hour=nrow(date_times))

# res<- multiApply::Apply(list(hour_vector,model), target_dim = list(NULL, c("lat", "lon")), #margins = 'hour',
#                          fun = UK_hourly,
#                          ncores = 2)


res<- multiApply::Apply(list(hour_vector),
                        margins = 'hour',
                        fun = UK_hourly,
                        ncores = 2)

dim(res$output1)
print(res)
# parallel::detectCores()

############################################################
#                Function:   UK-hourly                     #
############################################################
UK_hourly <- function(hour) {
  
  print(paste0("loop UK number:",hour))
  
  df_obs_hour   <- df_obs_sub[hour,-1]
  ok_hour <- ok_total[hour,]
  llista_uk                     <- data.frame(Y=local_stations@coords[,2],
                                           X=local_stations@coords[,1],
                                           NO2_obs=as.numeric(df_obs_hour[1,]),
                                           NO2_model=as.numeric(ok_hour[1,]),
                                           code=colnames(df_obs_hour))
  
  llista_uk                     <- na.omit(llista_uk)
  llista_uk[llista_uk$NO2_obs<0,"NO2_obs"] <- NA
  llista_uk                     <- na.omit(llista_uk)
  
  coordinates(llista_uk)  <- ~X+Y
  sp::proj4string(llista_uk) <- crs_model_utm
  
  #NO2.fit_uk <- automap::autofitVariogram(formula=NO2_obs~NO2_model,input_data=llista_uk, verbose=TRUE)
  NO2.fit_uk=readRDS(variogram_uk)
  
  loocv_h              <- krige.cv(NO2_obs~NO2_model,llista_uk,NO2.fit_uk$var_model)
  uk_loocv_h           <- matrix(data=loocv_h@data$var1.pred, nrow=1)
  #uk_loocv_h           <- matrix(NA,ncol=nrow(llista_uk), nrow=1)
  colnames(uk_loocv_h) <- llista_uk$code
  
  LOOCV_parallel <-  plyr::rbind.fill(as.data.frame(LOOCV_total_1row), as.data.frame(uk_loocv_h))
  
  return(as.numeric(LOOCV_parallel))
}