library(urbanR)
library(gstat)
library(sp)
library(automap)


sim_folder_path       <- "/esarchive/scratch/jmateu/Codes/run_geostats/"
local_stations       <- readRDS(file = paste0(sim_folder_path,"local_stations.rds")) 
date_times       <- readRDS(file = paste0(sim_folder_path,"date_times.rds")) 
df_obs_sub   <- readRDS(file = paste0(sim_folder_path,"df_obs_sub.rds")) 
ok_total <- readRDS(file = paste0(sim_folder_path,"ok_total.rds")) 
crs_model_utm         <- "+proj=utm +zone=31 +ellps=intl +units=m +no_defs"
variogram_uk <- "/esarchive/scratch/mhajji/Meriem/UK_regional/Variogram_UK.rds"

############################################################
#                                                          #
#                 Loop Universal Kriging                   #
#                                                          #
############################################################

print("LOOCV")
LOOCV_total_1row           <- matrix(NA,ncol=nrow(local_stations@data),nrow=0)
colnames(LOOCV_total_1row) <- local_stations$ref_obs 

#
hour_vector <- 1:nrow(date_times)
dim(hour_vector) <- c(hour=nrow(date_times))

force_working = FALSE

res<- multiApply::Apply(list(hour_vector),
                        margins = 'hour',
                        fun = UK_hourly,
                        ncores = 2)

dim(res$output1)
print(res)
# parallel::detectCores()

############################################################
#                Function:   UK-hourly                     #
############################################################
UK_hourly <- function(hour) {
  
  print(paste0("loop UK number:",hour))
  
  df_obs_hour   <- df_obs_sub[hour,-1]
  ok_hour <- ok_total[hour,]
  llista_uk                     <- data.frame(Y=local_stations@coords[,2],
                                           X=local_stations@coords[,1],
                                           NO2_obs=as.numeric(df_obs_hour[1,]),
                                           NO2_model=as.numeric(ok_hour[1,]),
                                           code=colnames(df_obs_hour))
  llista_uk                     <- na.omit(llista_uk)
  llista_uk[llista_uk$NO2_obs<0,"NO2_obs"] <- NA
  llista_uk                     <- na.omit(llista_uk)
  coordinates(llista_uk)  <- ~X+Y
  sp::proj4string(llista_uk) <- crs_model_utm
  
  if (force_working) {
    NO2.fit_uk=readRDS(variogram_uk)
    uk_loocv_h           <- matrix(NA,ncol=nrow(llista_uk), nrow=1)
  } else {
    NO2.fit_uk <- automap::autofitVariogram(formula=NO2_obs~NO2_model,input_data=llista_uk, verbose=TRUE)
    loocv_h              <- gstat::krige.cv(NO2_obs~NO2_model,llista_uk,NO2.fit_uk$var_model)
    uk_loocv_h           <- matrix(data=loocv_h@data$var1.pred, nrow=1)
  }

  colnames(uk_loocv_h) <- llista_uk$code
  LOOCV_parallel <-  plyr::rbind.fill(as.data.frame(LOOCV_total_1row), as.data.frame(uk_loocv_h))
  return(as.numeric(LOOCV_parallel))
}
