# Substitue all NA's in df_obs_sub by gbm predictions based on ok_total.

# Needs packages:
library(urbanR)
library(gbm)
library(lubridate)
library(caret)

# TO DO: 
# Extend period:
# date_time_begin_hist <- c(2019,01,01,00)
# date_time_end_hist   <- c(2021,01,01,00)


obs_filled <- df_obs_sub 
 
for (station in local_stations$ref_obs)
  # station <- local_stations$ref_obs[1]
  
  Test_results_station  <-  data.frame(date=date_times , Test_results = NA)
  Obs_Test_station      <-  data.frame(date=date_times , Obs_Test = NA)
  Obs_Train_station     <-  data.frame(date=date_times , Obs_Train = NA)
  
  # Select station data
  single_station_data <- data.frame(date            = df_obs_sub$date,
                                    obs             = df_obs_sub[,station],
                                    model           = ok_total[,station],
                                    hour_of_the_day = hour(df_obs_sub$date),
                                    day_of_year     = yday(df_obs_sub$date),
                                    day_of_week     = wday(df_obs_sub$date))
  
  single_station_data                     <- na.omit(single_station_data)

  # Split train/test fractions 
  train.fraction <- 0.90
  smp_size    <- floor(train.fraction * nrow(single_station_data))
  
# train_ind   <- sample(seq_len(nrow(single_station_data)), size = smp_size)
  # data_training = single_station_data[train_ind,]
  # data_test = single_station_data[-train_ind,]

    test_ind <- 140 : (140 + 48)
    data_training = single_station_data[-test_ind,]
    data_test = single_station_data[test_ind,]

  
  
  # Preliminary fit
  n.trees = 5000 
  Vx_formula <- "obs ~ model + hour_of_the_day + day_of_year + day_of_week"
  gbm1 <- gbm(as.formula(Vx_formula),
              data = data_training,
              distribution = "gaussian",
              n.trees = n.trees,
              interaction.depth = 1,
              shrinkage = 0.001,
              cv.folds = 5,
              n.cores = NULL, # will use all cores by default
              verbose = FALSE
  )  
  gbm.fit<- gbm1

  # print results
  print(gbm.fit)
  
  # plot loss function as a result of n trees added to the ensemble
  gbm.perf(gbm.fit, method = "cv")
  
  # get MSE 
  sqrt(min(gbm.fit$cv.error))
  
  # Variable importance
  gbm.fit.final <- gbm.fit
  par(mar = c(5, 8, 1, 1))
  summary(
    gbm.fit.final, 
    cBars = 10,
    method = relative.influence, # also can use permutation.test.gbm
    las = 2
  )
  par(mar = c(5, 8, 1, 1))
  summary(
    gbm.fit.final, 
    cBars = 10,
    method = relative.influence, # also can use permutation.test.gbm
    las = 2
  )
  
  res_training <- predict( gbm1 , newdata=data_training)
  res_test <- predict( gbm1 , newdata=data_test[,c("model","hour_of_the_day","day_of_year","day_of_week")])
  
  Test_results <- data.frame(date = data_test[,'date'], Test_results=res_test )
  Test_results_station[Test_results_station$date%in%Test_results$date ,"Test_results"] <- Test_results[,"Test_results"]
  
  
  Obs_Test <- data.frame(date = data_test[,'date'], Obs_Test=data_test$obs )
  Obs_Test_station[Obs_Test_station$date%in%Obs_Test$date ,"Obs_Test"] <- Obs_Test[,"Obs_Test"]
  
  Obs_Train <- data.frame(date = data_training[,'date'], Obs_Train=data_training$obs )
  Obs_Train_station[Obs_Train_station$date%in%Obs_Train$date ,"Obs_Train"] <- Obs_Train[,"Obs_Train"]
  
  
  data_single     <- data.frame(date                    = date_times,
                                Obs_Train_station       = Obs_Train_station$Obs_Train,
                                Obs_Test_station        = Obs_Test_station$Obs_Test,
                                Test_results_station    = Test_results_station$Test_results,
                                model                   = ok_total[,station])
  
  
  mm <- "January"
    # TIME_PLOT
    png(file = paste0(save_time_series,local_stations$ref_obs[l],"_", local_stations$station_name[l],"_",mm,".png"), width = 800, height = 400)
    openair::timePlot(selectByDate(data_single, year = 2021, month = mm),
                      pollutant = c("Obs_Train_station","Obs_Test_station","Test_results_station","model"),
                      group = TRUE,
                      title=station,
                      y.relation = "free",
                      ylab="NO2 (ug/m3)",
                      cols = c("black", "blue", "grey"),
                      lty = c("dotted","solid","dashed"),
                      lwd = c(3,2,2), 
                      name.pol = c("Obs Train station","Obs Test station", "Test results station", "Model"))
    dev.off()
  
  test_performace <- postResample(pred = res_test, obs = data_test$obs)
  training_performace <- postResample(pred = res_training, obs = data_training$obs)
  

